/*
    fi32 -- File directory listing utility with the ability to parse
	parts of the directory and create custom scripts.

    Copyright (C) 2005, 2007  Norm London, London Computing

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	norm_london@londoncomputing.com

    Compile with Visual Studio 6 or later.  
	cl fi32.c 

*/


#include <windows.h>

#define MAIN
#include <stdio.h>
#include <stdlib.h>
#include <io.h>
char drive[_MAX_DRIVE], dir[_MAX_DIR], fname[_MAX_FNAME], ext[_MAX_EXT];
char dir32[_MAX_DIR], fname32[_MAX_FNAME], ext32[_MAX_EXT];



typedef struct _FTIME {    /* ftime */
    unsigned twosecs : 5;
    unsigned minutes : 6;
    unsigned hours   : 5;
} FTIME;
typedef struct _FDATE {    /* fdate */
    unsigned day     : 5;
    unsigned month   : 4;
    unsigned year    : 7;
} FDATE;

#include <direct.h>

#define CPMEOF 26
#define CHDIR 0x3b

#include <string.h>
#include <ctype.h>
#include <process.h>
#include <stdlib.h>
#include <time.h>

FILE *out; /* number of output device */
FILE *in_file = stdin;         /* default to read stdin */
/* flags set by  set_args() */
int firstarg,prompt_filename,total_flag,totals_only,dflag,attrib,nflag,aflag;
int dflagOnly,fflag;
char spaces[_MAX_DIR];

char *comma_delim(_int64 lno,char *buffer,char mb,char kb) {
/* returns the long integer in comma delimited format */
/* Assume buffer is long enough to hold the largest value */
/* mb = 1; return result in MB; kb = 1 return result in KB */
	char b1[36],*b; /* should be long enough to hold largest ltoa value */
	int i,n,l,k,c;
	_int64 no;
	no = lno;
	if (mb) {
		no = (lno / (1024*1024));
		if (no > -1.0 && no < 1.0) {
			kb = 1;
			mb = 0;
		}
	}

	if (kb) {
		no = (lno / 1024);
		if (no > -1.0 && no < 1.0) {
			no = lno;
			kb = 0;
		}
	}

	sprintf(b1,"%16I64d",no);
	for (b=b1;*b==' ';++b) ;  // Remove leading blanks

	l = strlen(b); /* l is length of string and index to 0 at end of string */
	n = l/3;	   /* n is number of commas to be inserted */
	if ((l % 3) == 0)
		--n;
	k = l + n;	   /* k is length of new string */
	buffer[k--] = b[l--];
	for (c=0,i=l; i >= 0; --i,--k) {
		buffer[k] = b[i];
		if (++c == 3 && i > 0) {
			buffer[--k] = ',';
			c = 0;
		}
	}

	if (mb)
		strcat(buffer,"MB");
	else if (kb)
		strcat(buffer,"KB");

	return buffer;
}



/* -- */
void fitimtoa(char asctime[8],FTIME *filetime) {
	char temp[9];
	/* convert to ascii and store as hh:mm:ss */
	sprintf(temp,"%2d:%2d:%2d",filetime->hours,filetime->minutes,filetime->twosecs+filetime->twosecs);
	strncpy(asctime,temp,8);
	if (asctime[0] == ' ') asctime[0] = '0';
	if (asctime[3] == ' ') asctime[3] = '0';
	if (asctime[6] == ' ') asctime[6] = '0';
	
}

void fidtetoa(char ascdate[8],FDATE *filedate,char *fmt) {
	char temp[9];
	/* convert to ascii and store as mm/dd/yy */
	if (*fmt == 'M') {
		sprintf(temp,"%2d/%2d/%2d",filedate->month,filedate->day,(filedate->year+1980)%100);
		strncpy(ascdate,temp,8);
		if (ascdate[0]==' ') ascdate[0] = '0';
		if (ascdate[3]==' ') ascdate[3] = '0';
		if (ascdate[6]==' ') ascdate[6] = '0';
	}
	else {
		sprintf(temp,"%4d%2d%2d",1900+(filedate->year+1980)%100,filedate->month,filedate->day);
		strncpy(ascdate,temp,8);
		if (ascdate[0]==' ') ascdate[0] = '0';
		if (ascdate[4]==' ') ascdate[4] = '0';
		if (ascdate[6]==' ') ascdate[6] = '0';
	}
}

#define LMAX	512
#define PMAX	512
#define ANY	    '?'
#define CLASS	'['
#define ENDCLASS ']'
#define ESCAPE  '\\'
#define STAR	'*'
#define RANGE '-'
#define ENDPAT  0
char	lbuf[LMAX];
char	pbuf[PMAX];
char	*pp;
uxmatch(char *line, char *pattern)

/* (partial) line to match	   */
/* (partial) pattern to match	 */
{
	register char   *l;	    /* Current line pointer	      */
	register char   *p;	    /* Current pattern pointer      */
	int		   op;	      	/* Pattern operation	      */
	char inclass = 0, lowvalue = 0, foundmatch = 0, escape = 0;
	l = line;
	p = pattern;
	if (*p == ENDPAT)
		return 1;
	
	while ((op = *p++) != ENDPAT) {
		escape = 0;
		if (op == '\\') {
			op = *p++;
			if (op)		
				escape = 1;
			else {	/* pattern ends with \. look for \ */
				--p;
				op = '\\';
			}
		}
		
		if (inclass) {  
			/* currently scanning a class spec [abc-de] */
			if (!foundmatch) {
				/* and we haven't found a match yet */
				if (lowvalue) {
					/* but we are in a range c-e */
					/* op is high value of range */
					if (tolower(*l) >= lowvalue && tolower(*l) <= tolower(op)) {
						/* we found a match */
						foundmatch = 1;
						lowvalue = 0;
						continue;
					}
					/* no match found, return failure */
					lowvalue = 0;
				}
				else {
					/* not in a range */
					if (*p == RANGE)  {
						/* in a range, op is low value */
						lowvalue = (char) tolower(op);
						++p;
						continue;
					}
					else
						lowvalue = 0;
					
					if (tolower(*l) == tolower(op)) {
						/* found match, whoopie */
						foundmatch = 1;
						continue;
					}
				}
			}
			if (op == ENDCLASS && !escape) {
				if (!foundmatch)					
					return 0;		/* not found, sorry charlie */
				inclass = 0;
				++l;
			}
		}
		else {
			switch (op) {
			case ANY:
				if (*l++ == '\0')
					return(0);
				break;
			case CLASS:
				inclass = 1;
				lowvalue = 0;
				foundmatch = 0;
				break;				
			case STAR:		 /* Zero or more characters	*/
				/* match line against rest of pattern */
				if (*l == 0)
					return 1;
				while (*l)
					if (uxmatch(l++,p))
						return 1;
					return 0;		 /* Nothing else worked */
			default:
				if (tolower(*l) != tolower(op))
					return(0);
				l++;
				break;
			}
		}
	}
	return *l == 0 ? 1 : 0;   /* if everything matches and at end of string, ok */
}


char *leadspc(char *s)
{
	char *t;
	t=s;
	while (*s == ' ' || *s == '0') *(s++) =' ';
	return t;
}

void nputs(char *str,int len,int maxlen)
{
	int i;
	int ln;
	/* put out first n characters of str.  If len is 0, put entire string */
	/* if len not 0, put out len to a maximum of maxlen */
	ln = strlen(str);
	if (len == 0)
		fwrite(str,1,ln,out);
	else {
		len = min(len,maxlen);
		i = min((unsigned)len,strlen(str));
		fwrite(str,1,i,out); /* write out first i positions */
		if (len > i)
			fwrite(spaces,1,len-i,out);
	}
}

void put_help()
{
	FILE *help;
	char helpname[_MAX_PATH];
	char cmd[_MAX_PATH];
	strcpy(helpname,getenv("TEMP"));
	if (*helpname == 0)
		strcpy(helpname,"C:");
	strcat(helpname,"\\FI32.HLP");
	help = fopen(helpname,"w");
	if (help == NULL) {
		fprintf(stderr,"Unable to create help file.\n");
		exit(0);
	}
	strcpy(cmd,"NOTEPAD ");
	strcat(cmd,helpname);
	
	fprintf(help,"FILES (Ver %s) File Display Utility for Windows NT and Windows 9x.\n",__DATE__);
	fprintf(help,"Copyright (c) Norm London, London Computing, 2005,2007.\n");

    fprintf(help,"This program is free software; you can redistribute it\n");
	fprintf(help,"and/or modify it under the terms of the GNU General Public License\n");
    fprintf(help,"as published by the Free Software Foundation; either version 2\n");
    fprintf(help,"of the License, or (at your option) any later version.\n");
	fprintf(help,"\n");
    fprintf(help,"This program is distributed in the hope that it will be useful,\n");
    fprintf(help,"but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
    fprintf(help,"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
    fprintf(help,"GNU General Public License for more details.\n");
	fprintf(help,"\n");
    fprintf(help,"You should have received a copy of the GNU General Public License\n");
	fprintf(help,"along with this program; if not, write to:\n");
    fprintf(help,"Free Software Foundation, Inc.\n59 Temple Place, Suite 330\nBoston, MA  02111-1307  USA\n");
	fprintf(help,"\n");
	fprintf(help,"norm_london@londoncomputing.com\n");
	fprintf(help,"\n");
    fprintf(help,"Compile with Visual Studio 6 or later.  \n");
	fprintf(help,"cl fi32.c \n\n");


	fputs("Usage is:\n",help);
	fputs("files [flags] [d:][\\path\\]filename[.ext]... [/format]-or-\n",help);
	fputs("files -C[infi] [flags] [/format]\n\n",help);
	fputs("where\td is drive letter\n\tfilename and ext may contain Unix Wildcards *, ? and [...]\n",help);
	fputs("\tpath is a file path.\n",help);
	fputs("\tFlags are:\n\t\t[-%] [-Ootfi] [-T[O]] [-D[O][A]] [-A[D[O]]] [-N] [-Bn] [-s]\n",help);
	fputs("\t-C[in]\tcauses a list of files to be read from the keyboard or \n",help);
	fputs("\t\tfile in.  The list must end with a null line.\n",help);
	fputs("\t-Ootfi\tcauses listing to go to otfi instead of the terminal\n",help);
	fputs("\t-T\tprints totals after each specification.\n",help);
	fputs("\t-TO\tprints totals ONLY\n",help);
	fputs("\t-Bn\tAssume each block in the file is n characters long when \n",help);
	fputs("\t\tcalculating total space used.  Set to 512 for HPFS, NTFS.\n",help);
	fputs("\t-N\tOmit CR-LF at end of each line. \n",help);
	fputs("\t-A\tShow all files, including Hidden/System files.\n",help); 
	fputs("\t\tIf omitted, show NORMAL files only.\n",help);
	fputs("\t-AD\tAll files and Directories, including Hidden/System.\n",help); 
	fputs("\t-ADO\tAll directories, including Hidden/System.\n",help);
	fputs("\t-D\tShow files and directories; else only show files.\n",help);
	fputs("\t-DA\tShow all files and directories, including hidden/system.\n",help);
	fputs("\t-DO\tDirectories ONLY (-A, show hidden/system directories also).\n",help);
	fputs("\t-DOA\tShow all directories, including hidden/system.\n",help);
	fputs("\t-%\tIf specified, use % instead of $ in the display format.\n\n",help);
	fputs("\t-s\tDisplay all named streams defined in the file.\n",help);
	fputs("\t\tThe normal contents of the file constiture an unnamed stream.\n",help);
	fputs("\t\tThis stream is always present. On Macintosh volumes the unnamed\n",help);
	fputs("\t\tstream contains the data fork. In addition, on any volume\n",help);
	fputs("\t\tthat supports the Macintosh, there is a second stream that is\n",help);
	fputs("\t\talways present: AFP_AFPIndex.  Macintosh files may also have\n",help);
	fputs("\t\ta second named stream: AFP_Resource, which contains the Macintosh\n",help);
	fputs("\t\tresource fork. Note:  Use of the -s option will cause FI32 to run\n",help);
	fputs("\t\tslowly because it is necessary to open every file in order to find\n",help);
	fputs("\t\tthe file stream information.\n",help);
    //fputs("\nPress RETURN to continue ...",help);
    //gtchar();
	fputs("\tFormat defines file display format.  Within format, use:\n",help);
	fputs("\t$nF for short filename. n is optional field width (1-8)\n",help);
	fputs("\t$nE for file extention (from 8.3 name) n is optional field width.\n\t\tIncludes leading period.\n",help);
	fputs("\t$ne for file extention (from 8.3 name. n is optional field width.\n\t\tDoes not include leading periond.\n",help);
	fputs("\t$np for file path using 8.3 conventions.  n is optional field width\n\t\t (1-255).  Includes leading,trailing \\.\n",help);
	fputs("\t$nP for file path using long file name conventions. n is optional field\n\t\t width (1-255).  Includes leading,trailing \\.\n",help);
	fputs("\t$nD for Drive letter.  n is optional field width (1-2).\n\t\tIncludes trailing :.\n",help);
	fputs("\t$nX for long filename. n is optional field width (1-255)\n",help);
	fputs("\t$N  -- LF, $G -- >, $L -- <, $B -- |, $$ -- $,  $- for -.\n",help);
	fputs("\t$nS for file size.  n is optional field width (1-12).\n",help);
	fputs("\t$nM for file mode.  n is optional field width (1-8).\n",help);
	fputs("\t\tWithin Mode, A -- Archive, D -- Directory\n\t\tH -- Hidden, S -- System, R -- ReadOnly\n",help);
	fputs("\t$nY  for file modification time/date stamp in YMD format.\n",help);
	fputs("\t$nYm for file modification time/date stamp in YMD format.\n",help);
	fputs("\t$nYc for file creation time/date stamp in YMD format.\n",help);
	fputs("\t$nYa for last file access time/date stamp in YMD format.\n",help);    
	fputs("\t\tIf n is 14 or 17, use a 2 digit year, else use 4 digit year.\n",help);	
	fputs("\t\t$nY YMD format is suitable for sorting.\n",help);
	fputs("\t$n<space> for n spaces. (E.G. $5 -- 5 spaces)\n",help);
	fputs("\tAll other characters print as is.\n",help);
	fputs("\nThis program has been placed in the public domain.\n",help);
	fclose(help);
	system(cmd);
	exit(0);
}


getname(int argc,char *argv[],char *ambname,unsigned short ambsize,
		int firstarg,int prompt_filename)
{
	int rc,l;
	static int i=0;
	if (i == 0) i = firstarg;
	if (prompt_filename) {
		if (fgets(ambname,_MAX_PATH,in_file) == NULL)
			rc = 0;
		else {
			rc = 1;
			for (l=strlen(ambname)-1;ambname[l]=='\r'||ambname[l]=='\n';--l)
				ambname[l] = 0;
		}
	}
	else {
		while (i < argc && argv[i][0] == '-') ++i;
		if ((rc = (i >= argc) || (argv[i][0] == '/') ? 0 : 1) != 0)
			strcpy(ambname,argv[i++]);
	}
	return rc;
}


void 
set_args(int argc, char *argv[], char *otfi, int *firstarg,
		 int *lastarg,int *prompt_filename,int *total_flag,
		 int *totals_only,char *ctl,int *use_blocksize, int *showStreams)
{
	int i;
    aflag = nflag = dflag = dflagOnly = fflag = 0;
	*lastarg = 0;
	*use_blocksize = 0;
	*prompt_filename = 0;
	*totals_only = 0;
	*total_flag = 0;
	*firstarg = 0;
	*otfi = 0;
	for (i = 1; i < argc; ++i)
		if (argv[i][0] == '-')
			switch(toupper(argv[i][1])) {
			case 'S':
				*showStreams = 1;
				break;
			case 'B':
				*use_blocksize = atol(argv[i]+2);
				break;
			case '%':
				*ctl = '%';
				break;
            case 'A': /* All files and Directories */
                aflag = 1;
				if (toupper(argv[i][2]) == 'D') {
					dflag = 1;
					if (toupper(argv[i][3]) == 'O')
						dflagOnly = 1;
				}
                break;
			case 'N': /* No linefeed */
				nflag = 1;
				break;
			case 'D': /* directories only */
				dflag = 1;
				if (toupper(argv[i][2]) == 'O') {
					dflagOnly = 1;
					if (toupper(argv[i][3]) == 'A') 
						aflag = 1;
				}
				else 
					if (toupper(argv[i][2]) == 'A')
						aflag = 1;
					break;
			case 'O': /* send output to a file */
				strcpy(otfi,argv[i]+2);
				strupr(otfi);
				fprintf(stderr,"Writing to: %s\n",otfi);
				break;
			case 'C': /* accept file list from console */
				*prompt_filename = 1;
				if (argv[i][2]) { /* redirected input file */
					in_file = fopen(argv[i]+2,"r");
					if (in_file <= 0) {
						in_file = 0;
						fprintf(stderr,"Unable to open %s\n",argv[i]+2);
						exit(1);
					}
				}
				break;
			case 'T': /* produce total for files found */
				*total_flag = 1;
				*totals_only = (toupper(argv[i][2]) == 'O');
				break;
			case 'F': /* Full Info - Print out Maximal Info */
				*total_flag = 1;
				aflag = 1;
				fflag = 1;
				dflag = 1;
				break;
			default:
				fprintf(stderr,argv[i]);
				fprintf(stderr," <=== I don\'t understand!\n");
				break;
		}
		else {
			if (! *firstarg)
				*firstarg = i;
			*lastarg = i;
		}
}


static char driveletter[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
void buildpath(char *path,unsigned pathsi,char *pathdir,unsigned pathdirsi)
{
	char *p;
	USHORT d;
	int l,i;
	USHORT pathsize= pathdirsi-3;
	long diskmap = 0;
	pathdir[1]=' ';
	pathdir[2]=0;
	pathdir[3]=0;
	pathdir[4]=0;
	if (path[1]==':') {
		strncpy(pathdir,path,2);
		*pathdir=(char) toupper(*pathdir);
		p = path+2;
		d = *pathdir-'@';
		l = 2;
	}
	else {
		p=path;
		d = _getdrive();
		pathdir[0]=driveletter[d-1];
		pathdir[1]=':';
		l = 2;
	}
    if (p[0] != '\\' && p[0] != '/') {
		pathdir[2]='\\';
		_getdcwd(d,pathdir,pathsize);
		l = strlen(pathdir);
	}
	
    for (i=strlen(p);i >= 0 && p[i] != '\\' && p[i] != '/'; --i) ;
	if (i > 0) {
		/* subdirectory specified */
        if (*p != '\\' && *p != '/')
            pathdir[l++] = '\\';
		strncpy(pathdir+l,p,i);
		pathdir[i+l]=0;
	}
	else if (i==0) {/* root directory specified */
		pathdir[2] = '\\';
		pathdir[3]=0;
	}
	strupr(pathdir);
}

void parsename(char *filename,char *name,char *ext)
{
	int i;
	char *c;
	if (strcmp(filename,"..") == 0) {
		strcpy(name,"..");
		*ext = 0;
		return;
	}
	if (strcmp(filename,".") == 0) {
		strcpy(name,".");
		*ext = 0;
		return;
	}
	c = strchr(filename,'.');
	if (c == NULL)
		i = -1;
	else
		i =c - filename;
	if (i == -1) {
		strcpy(name,filename);
		ext[0]=0;
	}
	else {
		strncpy(name,filename,i);
		name[i]=0;
		strcpy(ext,filename+i+1);
	}
}


void print_totals(
	char *prevpath, 
	FILE *out, 
	int filekt,
	_int64 space_used,
	_int64 total_space,
	_int64 space_avail,
	char *actsize)
{
    if (prevpath[1] != ':' && strncmp(prevpath,"\\\\",2) != 0 && strncmp(prevpath,"//",2) != 0)  {
		putc(prevpath[0],out);
		nputs(":",0,1);
	}                       
	nputs(prevpath,0,1);
	nputs(" ",0,1);
	sprintf(actsize,"%d",filekt);
	nputs(actsize,0,1);
	nputs(" files, ",0,1);
	comma_delim(space_used,actsize,0,0);
//	sprintf(actsize,"%I64d",space_used);
	nputs(actsize,0,1);
    if (total_space != 0) {
        nputs(" used, ",0,1);
//      sprintf(actsize,"%I64d",space_avail);
		comma_delim(space_avail,actsize,0,0);
        nputs(actsize,0,1);
        nputs(" free of ",0,1);
 //     sprintf(actsize,"%I64d",total_space);
		comma_delim(total_space,actsize,0,0);
        nputs(actsize,0,1);
    }
	nputs(" bytes.\n",0,1);
}


char actname[_MAX_FNAME],actext[_MAX_EXT],command[255],otfi[_MAX_DIR];
char ambname[_MAX_DIR],actpath[_MAX_DIR],datetime[26],actsize[21];
char prevpath[_MAX_DIR],fi_mode[5],searchpath[_MAX_DIR],fullpath[_MAX_DIR];

/* Concatenate pth to searchpath and parse */
void catpath(char *pth) {
    strcat(searchpath,pth);
    strcpy(actpath,searchpath);
    _fullpath(searchpath,actpath,sizeof(actpath));
    _splitpath(searchpath,drive,dir,fname,ext);
    strcpy(actpath,drive);
    strcat(actpath,dir);
    strupr(actpath);
}

#define INVALID_HANDLE -1
/* Return 1 if path is a directory.  Use DosFindFirst to test */
/* Can not use stat because it does not handle UNC names properly */
int is_dir(char *path) {
    struct _finddata_t dtax;
	
    long hdir;
	
    char searchpath[_MAX_PATH];
    strcpy(searchpath,path);
    strcat(searchpath,"/*.*");
    hdir = INVALID_HANDLE;
	
    hdir = _findfirst(searchpath,&dtax);
	if (hdir == INVALID_HANDLE)
		return 0;
	_findclose(hdir);
	return 1;
}



int IsNetwork(int drive) {
	/* Returns 1 if a Network Drive, 0 if a Physical Drive */
	unsigned int rc;
	char driv[4]="C:\\";
	driv[0] = '@' + drive;
	rc = GetDriveType(driv);
	return (rc == DRIVE_REMOTE) ? 1 : 0;
}

int LongPathName(char *longname,char *shortname,char isdir) {
	HANDLE ffhandle;
	int l,i,kt;
	WIN32_FIND_DATA FindData;
	char lname[_MAX_PATH];
	char fname[_MAX_PATH];
	char drive[_MAX_DIR], dir[_MAX_DIR], name[_MAX_FNAME], ext[_MAX_EXT];

	_splitpath(shortname,drive,dir,name,ext);
	if (strlen(dir) > 1 && strncmp(dir,"\\\\",2) == 0) {
		// Have a UNC Name, Move to Drive
		for (i=0, kt=0; i < strlen(dir) && kt < 4; ++i) {
			drive[i] = dir[i];
			if (dir[i] == '\\' || dir[i] == '/')
				++kt;
		}
		// i should now point to the 1st position of the actual directory
		drive[i-1] = 0;
		strcpy(dir,dir+i-1);
	}
	
	if (*dir == 0 || strcmp(dir,"\\") == 0 || strcmp(dir,"/") == 0) {
		// get long version of fname by looking up in directory		
		strcpy(fname,drive);
		strcat(fname,dir);
		strcat(fname,name);
		strcat(fname,ext);
		ffhandle = FindFirstFile(fname,&FindData);
		if (ffhandle ==  INVALID_HANDLE_VALUE)
			return -1;
		// return longname in longname field);
		strcpy(longname,drive);
		strcat(longname,dir);
		strcat(longname,FindData.cFileName);
		if (isdir)
			strcat(longname,(dir[0]=='\\'? "\\" : "/"));
		FindClose(ffhandle);
		return 0;
	}
	else {
		// file is in a directory, get long name of directory
		strcpy(fname,drive);
		strcat(fname,dir);
		l = strlen(fname);
		if (l==2 && fname[0] == '\\' && fname[1] == '\\') {
			strcpy(lname,fname);
			return 0;
		}
		if (fname[l-1] == '/' || fname[l-1] == '\\')
			fname[l-1] = 0; //remove any trailing Slashes
		if (LongPathName(lname,fname,1) == -1)
			return -1;
		strcpy(fname,lname);
		strcat(fname,name);
		strcat(fname,ext);
		ffhandle = FindFirstFile(fname,&FindData);
		if (ffhandle ==  INVALID_HANDLE_VALUE)
			return -1;
		// return longname in longname field);
		strcpy(longname,lname);
		strcat(longname,FindData.cFileName);
		if (FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
			strcat(longname,(dir[0]=='\\'?"\\":"/"));
		FindClose(ffhandle);
		return 0;
	}
}

char *formatMDY(
	char *datetime,
	char *c,
	int width,
	time_t time_access, 
	time_t time_create, 
	time_t time_write) 
{

	struct tm *tm;
	
	switch(toupper(*(c+1))) {
	case 'A':
		if (time_access == -1L)
			tm = localtime(&time_write);
		else
			tm = localtime(&time_access);
		++c;
	    // Fat file systems return write time only																		
		break;
	case 'C':
		if (time_create == -1L)
			tm = localtime(&time_write);
		else
			tm = localtime(&time_create);
		++c;
		// Fat file systems return write time only																		
		break;
	case 'M':
		tm = localtime(&time_write);
		++c;
		break;
	default:
		tm = localtime(&time_write);
		break;
	}
	sprintf(datetime,"%2d/%2d/%4d %2d:%2d:%2d",
		tm->tm_mon+1, tm->tm_mday, tm->tm_year+1900,
		tm->tm_hour, tm->tm_min, tm->tm_sec);
	if (datetime[ 0] == ' ') datetime[ 0] = '0';
	if (datetime[ 3] == ' ') datetime[ 3] = '0';
	if (datetime[11] == ' ') datetime[11] = '0';
	if (datetime[14] == ' ') datetime[14] = '0';
	if (datetime[17] == ' ') datetime[17] = '0';
	if (width == 17 || width == 14) // Use a 2 digit year
		strcpy(datetime+6, datetime+8);
	return c;
}

char *formatYMD(
	char *datetime, 
	char *c, 
	int width,
	time_t time_access, 
	time_t time_create, 
	time_t time_write) 
{

	struct tm *tm;
	
	switch(toupper(*(c+1))) {
	case 'A':
		if (time_access == -1L)
			tm = localtime(&time_write);
		else
			tm = localtime(&time_access);
		++c;
		// Fat file systems return write time only																		
		break;

	case 'C':
		if (time_create == -1L)
			tm = localtime(&time_write);
		else
			tm = localtime(&time_create);
		++c;
		// Fat file systems return write time only																		
		break;

	case 'M':
		tm = localtime(&time_write);
			++c;
		break;
		
	default:
		tm = localtime(&time_write);
		break;
	}
	sprintf(datetime,"%4d/%2d/%2d %2d:%2d:%2d",
		tm->tm_year+1900, tm->tm_mon+1, tm->tm_mday, 
		tm->tm_hour, tm->tm_min, tm->tm_sec);
	if (datetime[ 5] == ' ') datetime[ 5] = '0';
	if (datetime[ 8] == ' ') datetime[ 8] = '0';
	if (datetime[11] == ' ') datetime[11] = '0';
	if (datetime[14] == ' ') datetime[14] = '0';
	if (datetime[17] == ' ') datetime[17] = '0';
	if (width == 17 || width == 14) // Use a two digit year
		strcpy(datetime,datetime+2);
	return c;
}

int FiMode(int w,unsigned attrib,char *fi_mode) {
	int width;
	width = w;
	if (width)
		width = min(width,5);
	else
		width = 5;
	strcpy(fi_mode,"     ");
								
	if (attrib & _A_RDONLY) fi_mode[3] = 'R';
	if (attrib & _A_HIDDEN) fi_mode[1] = 'H';
	if (attrib & _A_SYSTEM) fi_mode[1] = 'S';
	if (attrib & _A_ARCH  ) fi_mode[4] = 'A';
	if (attrib & _A_SUBDIR) fi_mode[0] = 'D';

	return width;
}

// Displays information about this file
void  WriteFileInfo(
	char *fname,
	char *ext,
	char *dir,
	char *c1,
	char ctl,					// Control character
	unsigned attrib,
	char *name,
	_int64 size,	
	time_t time_access,
	time_t time_create,
	time_t time_write,
	char totals_only,
	char *actpath, 
	FILE *out) 
{
	char oldc;
	char *c;
	char more=0;
	int i,k;
	int width;
	char buff[21];
					
	oldc = 0;
	c = c1;
		
	while (*c && !totals_only) {
		if ((*c == '\\' || *c == '/')
			&& oldc == 16 && strlen(actpath+2) == 1) {
			oldc = '\\';
			++c;
			continue;
		}

		if (*c != ctl) {
			putc((oldc = *c),out);
			++c;
			continue;
		}

		
		width = 0;
		more = 1;
		while (more) {
			more = 0;
			i = *(++c);
			switch(i) {
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':
				width = width*10 + *c - '0';
				more = 1;
				break;
									
			case 'M': /* %M is file mode */
			case 'm': 

				// Format the File Mode information in dtax.attrib
				width = FiMode(width,attrib,fi_mode);
				i = 0;
				if (aflag || dflag)
					nputs(fi_mode,width,5);
				else
					nputs(fi_mode+3,width,2);
				break;
									
			case 'N': /* %N used to indicate new line */
			case 'n':
				putc('\n',out);
				oldc = '\n';
				break;
								
			case 'S': /* file size, use commas */
				if (width)
					width = min(width,20);                                                                           
				else
					width = 20;
				for (k=0;k<20;++k)
					actsize[k] = ' ';
				actsize[20] = 0;
				comma_delim(size,buff,0,0);
				if (strlen(buff) > 20)
					strcpy(actsize,buff+strlen(buff)-20);
				else
					strcpy(actsize+20-strlen(buff),buff);
							
				nputs(actsize+20-width,0,width);
				oldc = 0;
				break;
			case 's': /* file size, no commas */
				if (width)
					width = min(width,12);                                                                           
				else
					width = 12;
								
				sprintf(actsize,"%12I64d",size);
				nputs(actsize+12-width,0,width);
				oldc = 0;
				break;
									
			case 'T': /* date, time stamp */
			case 't':
				/* Use Create, Access or Mod time? */
				c = formatMDY(datetime,c,width,time_access,time_create,time_write);
				nputs(datetime,width,19);
				oldc = 0;
				break;
									
			case 'Y': /* date, time stamp in YMD format */
			case 'y':
				/* Use Create, Access or Mod time? */
				c=formatYMD(datetime,c,width,time_access,time_create,time_write);
				nputs(datetime,width,19);
				oldc = 0;
				break;
										
			case 'D': /* volume */
			case 'd':
				nputs(drive,width,_MAX_DRIVE);
				oldc = '\4';
				break;
										
			case 'X': /* Extended (long) file name */
			case 'x':
				nputs(name,width,_MAX_FNAME);
				oldc = 0;
				break;
											
			case 'F': /* file name */
			case 'f':
				/* precede filename with \ if prev output was path and not root */
				nputs(fname,width,_MAX_FNAME);
				oldc = '\6';
				break;
											
			case 'E': /* file extension */
				nputs(ext,width,_MAX_EXT);
				oldc = '\5';
				break;

			case 'e': // file extension, no period
				nputs(ext+1,width,_MAX_EXT);
				oldc = '\5';
				break;
											
			case 'P': 
			case 'p':
				/* file path; precede filepath with \ if previous character was : */
				oldc = 16; /* set to control P */
				nputs(i=='p' ? dir : dir32,width,_MAX_DIR);
				break;
											
			case 'G': /* > sign */
			case 'g':
				oldc = 0;
				nputs(">",1,1); 
				break;
											
			case 'L': /* > sign */
			case 'l':
				oldc = 0;
				nputs("<",1,1); 
				break;
											
			case 'B': /* | */
			case 'b':
				oldc = 0;
				nputs("|",1,1);
				break;
										
			case ' ': /* space */
				oldc = ' ';
				if (width)
					nputs(spaces,width,width);
				else
					putc(' ',out);
				break;
											
			default:
				putc((oldc = *c),out);
			} /* end switch *c */
		} /* end while more */
		++c;
	} //end while *c ...
}

// check the file name and attribs 
// return 1 to skip the file else 0;
int skip_this_file(char *name, unsigned attrib,char *ambname,char unix_wildcards) {
	int l;
	// skip . , .. entries
	if (name[0] == '.')
		return 1;
	// skip hidden , system unless -A specified
	if (attrib & (_A_HIDDEN | _A_SYSTEM))
		if (!aflag)
			return 1;
	// skip sub-directories unless -D specified
	if (attrib & _A_SUBDIR)
		if (!dflag)
			return 1;
	// skip normal files if -DO specifield
	if (dflagOnly && (attrib & _A_SUBDIR) == 0)
		return 1;
	// If we are using unix wildcards, check name				
	if (unix_wildcards) {
		l = strlen(name);
		if (strchr(name,'.') == NULL) {
			name[l] = '.';
			name[l+1] = 0;
		}
		if (! uxmatch(name,ambname))
			return 1;
		name[l] = 0;
	}
	return 0;
}


void parse_name_for_display(
	char *name,
	char *full_path,
	char *short_name)
{
	int rc1;
	int short_name_len = _MAX_PATH;
	strcpy(searchpath,actpath);
	strcat(searchpath,name);
	rc1 = LongPathName(full_path,searchpath,0);

	_splitpath(full_path,drive,dir32,fname32,ext32);
			
	rc1 = GetShortPathName(searchpath,short_name,short_name_len);
	if (rc1 == 0) {
		strcpy(short_name,"Not Supported");
		strcpy(dir,dir32);
		strcpy(fname, fname32);
		if (strlen(fname) > 8) {
			strcpy(ext,"....");
			fname[8] = 0;
		}
		else 
			strcpy(ext,ext32);
		if (strlen(ext32) > 4)
			strcpy(ext,"....");
	}
	else
		_splitpath(short_name,drive,dir,fname,ext);

	if (strcmp(fname,".") == 0 && strcmp(ext,".") == 0) {
		strcpy(fname,"..");
		ext[0] = 0;
	}
	else if (fname[0] == 0 && strcmp(ext,".") == 0) {
		strcpy(fname,".");
		ext[0] = 0;
	}
}


_int64 calculate_filesize(_int64 size,int use_blocksize) {
	_int64 filesize;
	filesize = (size / use_blocksize) * (long) use_blocksize;
	if (size > (unsigned _int64)filesize)
		filesize += (_int64) use_blocksize;
	return filesize;
}

WIN32_STREAM_ID stream;
VOID *context;
HANDLE fileHandle;



// Close the File Handle and shut down Backup Read
int CloseBackupReadHandle(WCHAR *streamname) {
	DWORD BytesRead;
	stream.dwStreamAttributes=0;
	stream.dwStreamId=0;
	stream.dwStreamNameSize=0;
	stream.Size.QuadPart = 0;
	streamname[0] = 0;
	BackupRead(fileHandle,(unsigned char *)&stream,0,&BytesRead,TRUE,FALSE,&context);
	CloseHandle(fileHandle);
	return 0;
}


int FirstStream(
	char *filename,
	char *StreamNameAscii,
	_int64 *filesize,
	int *streamId)
{

	WCHAR streamname[_MAX_PATH];
	DWORD BytesRead,actualLow,actualHigh,namesize;
	DWORD rc;
	int StreamHeaderSize = ((LPBYTE)stream.cStreamName - (LPBYTE)&stream);

	context=NULL;
	fileHandle = CreateFile(filename,GENERIC_READ,FILE_SHARE_READ,
			NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
	if (fileHandle == INVALID_HANDLE_VALUE) {
		rc = GetLastError();
		return 0;
	}
	rc = 1;
	BytesRead=1;
	stream.dwStreamAttributes=0;
	stream.dwStreamId=0;
	stream.dwStreamNameSize=0;
	stream.Size.QuadPart = 0;
	streamname[0] = 0;
	// Read the first 20 bytes of the file into stream
	// This gives the stream id, stream attributes,
	// 
	rc=BackupRead(fileHandle,(unsigned char *)&stream,StreamHeaderSize,&BytesRead,FALSE,FALSE,&context);
	if (rc && BytesRead) {
		namesize = stream.dwStreamNameSize;
		*filesize = stream.Size.QuadPart;
		*streamId = stream.dwStreamId;
		if (namesize) {
			// read the file name
			rc = BackupRead(fileHandle,(unsigned char *)streamname,namesize,
			&BytesRead,FALSE,FALSE,&context);
			namesize /= sizeof(WCHAR);
			streamname[namesize]=0;
		}
		
		sprintf(StreamNameAscii,"%ls",streamname);
		rc = BackupSeek(fileHandle,stream.Size.LowPart,stream.Size.HighPart,&actualLow,&actualHigh,&context);
		return 1;
	}
	
	CloseBackupReadHandle(streamname);
	
	return 0;
}

int NextStream(
	char *StreamNameAscii,
	_int64 *filesize,
	int *streamId)
{

	WCHAR streamname[_MAX_PATH];
	DWORD BytesRead,actualLow,actualHigh,namesize;
	DWORD rc;
	int StreamHeaderSize = ((LPBYTE)stream.cStreamName - (LPBYTE)&stream);

	rc = 1;
	BytesRead=1;
	stream.dwStreamAttributes=0;
	stream.dwStreamId=0;
	stream.dwStreamNameSize=0;
	stream.Size.QuadPart = 0;
	streamname[0] = 0;
	// Read the first 20 bytes of the file into stream
	// This gives the stream id, stream attributes,
	// 
	rc=BackupRead(fileHandle,(unsigned char *)&stream,StreamHeaderSize,&BytesRead,FALSE,FALSE,&context);
	if (rc && BytesRead) {
		namesize = stream.dwStreamNameSize;
		*filesize = stream.Size.QuadPart;
		*streamId = stream.dwStreamId;
		if (namesize) {
			// read the file name
			rc = BackupRead(fileHandle,(unsigned char *)streamname,namesize,
			&BytesRead,FALSE,FALSE,&context);
			namesize /= sizeof(WCHAR);
			streamname[namesize]=0;
		}
		
		sprintf(StreamNameAscii,"%ls",streamname);
		rc = BackupSeek(fileHandle,stream.Size.LowPart,stream.Size.HighPart,&actualLow,&actualHigh,&context);
		return 1;
	}
	CloseBackupReadHandle(streamname);
	return 0;
}


#define PATHLEN 255
void main(int argc,char *argv[])
{
	struct _finddatai64_t dtax;
	int lastarg,firsttime;
	char *c1,ctl='$',*defstr,*dirstr;
	static char defstrp[]="%8f%4e %20S %16y %m in %d%p %x";
	static char dirstrp[]="%8f%4e%m %y %d%p%x";
	static char defstrd[]="$8f$4e $20S $16y $m $d$P$x";
	static char dirstrd[]="$8f$4e $m $y $d$p$x";
	static char deffull[]="$8f$4e $20S $yc $ym $ya $m $d$P$x";
	static char fulltitle[]="File Name         File Size Create Date         Modification Date   Last Access Date    Mode  Full Path Name\n";

	char driv[4]="C:\\";
	char showStreams = 0;
	int l,i,rc,filekt;
	char short_name[_MAX_PATH];
	char full_path[_MAX_PATH];
	long hdir;
	
	unsigned sectors,size,units,blocksize,use_blocksize,sectors_avail;
	unsigned block_sizes[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	unsigned net_drive[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	_int64 space_used,total_space,filesize,space_avail;
	_int64 total_space_used,sum_total_space,total_filekt,longtemp=0;
	char unix_wildcards = 0, foundfile = 0;
	long sectors_per_cluster, bytes_per_sector,	free_clusters, total_clusters;
	
	
	total_space_used = 0;
	sum_total_space = 0;
	total_filekt = 0;
	filekt = 0;
	space_used = 0;
	_tzset();
	
	set_args(argc,argv,otfi,&firstarg,&lastarg,&prompt_filename,
		&total_flag,&totals_only,&ctl,&use_blocksize,&showStreams);
	
	if (fflag) {
		defstr = deffull;
		dirstr = deffull;
	}
	else {
		if (ctl == '$') {
			defstr = defstrd;
			dirstr = dirstrd;
		}
		else {
			defstr = defstrp;
			dirstr = dirstrp;
		}
	}
	
	*prevpath  = 0;
	firsttime = 1;
	if (firstarg ==0 && prompt_filename == 0)
		put_help();
	c1 = dflag ? dirstr : defstr;
	if (*otfi) {
		if ((out = fopen(otfi,"w")) == NULL) {
			fprintf(stderr,"%s <=== Create Error.\n",otfi);
			out = stdout;
		}
	}
	else
		out = stdout;

	if (fflag) 
		nputs(fulltitle,0,0);

    for (i=0; i < _MAX_DIR; spaces[i++] = ' ') ;   /* fill blanks */
	
	for (i=firstarg; i < argc && argv[i][0] != '/'; ++i) ; /* look for / */
	
	if (i < argc) { /* found one */
		c1 = command;
		strcpy(command,&argv[i][1]);
		while (++i <= lastarg) {
			strcat(command," ");
			strcat(command,argv[i]);
		}
	}
	
	while (getname(argc,argv,ambname,sizeof(ambname),firstarg,prompt_filename)) {
		/*		buildpath(ambname,sizeof(ambname),actpath,sizeof(actpath)); */
		_fullpath(searchpath,ambname,sizeof(ambname));
		_splitpath(searchpath,drive,dir,fname,ext);
		strcpy(actpath,drive);
		strcat(actpath,dir);
		strupr(actpath);
		if (strncmp(actpath,"\\\\",2)) {
			// Grab disk info
			i = actpath[0] - '@';
			if (block_sizes[i]) {
				blocksize = block_sizes[i];
				total_space = (long) units * (long) blocksize;
				space_avail = (long) sectors_avail * (long) blocksize;
				if (net_drive[i])
					if (blocksize > 8192)
						blocksize = 512;
					/* Use 512 for sector size of Network Drives with blocksize>8192*/
			}
			else {
				driv[0] = actpath[0];
				GetDiskFreeSpace(driv, &sectors_per_cluster, &bytes_per_sector,	&free_clusters, &total_clusters);
			}
			units = (unsigned) total_clusters;
			sectors = (unsigned) sectors_per_cluster;
			size = bytes_per_sector;
			sectors_avail = (unsigned) free_clusters;
			blocksize = sectors * size;
            block_sizes[i] = blocksize;
            total_space = (_int64) units * (_int64) blocksize;
            space_avail = (_int64) sectors_avail * (_int64) blocksize;
            if (IsNetwork(i)) {
                net_drive[i] = 1;
                if (blocksize > 8192)
                    blocksize = 512;
				/* Use 512 for sector size of Network Drives with blocksize>8192*/
			}
		}
		else { /* Universal naming convention, can't get info */
			units = 0;
			sectors = 1;
            size = 512;
			sectors_avail = 0;
            blocksize = 512;
            total_space = (long) units * (long) blocksize;
            space_avail = (long) sectors_avail * (long) blocksize;
        }
		
		if (use_blocksize == 0)
			use_blocksize = blocksize;
		
		sum_total_space += total_space;
		strupr(ambname);
		
		if (strchr(ambname,'*') == NULL && strchr(ambname,'[') == NULL) {
			strcpy(searchpath,ambname);
			l = strlen(searchpath) - 1;
            if (strchr("\\/:",searchpath[l]) != NULL)
                catpath("*.*");
			else
                if (searchpath[l] == '.' || is_dir(searchpath))
                    catpath("/*.*");
				
				unix_wildcards = 0;
		}
		else {
			strcpy(searchpath,actpath);
			if (searchpath[(l = strlen(searchpath)-1)] != '\\' &&
				searchpath[l] != '/')
				strcat(searchpath,"\\");
			strcat(searchpath,"*.*");
			unix_wildcards = 1;
		}

        if (strcmp(prevpath,actpath)) {
            if (total_flag && !firsttime)
                print_totals(prevpath,out,filekt,space_used,total_space,space_avail,actsize);
            space_used = 0;
            filekt = 0;
            strcpy(prevpath,actpath);
        }
        firsttime = 0;
		
		for (l=strlen(ambname)-1;
		l && ambname[l] != '\\' && ambname[l] != '/' && ambname[l] != ':';
		--l) ;
		if (l)
			strcpy(ambname,ambname+l+1);
		
		foundfile = 0;

		// Look through the directory for files that meet the criteria for display
		for (rc =0,hdir = _findfirsti64(searchpath,&dtax); 
		hdir != -1 && rc != -1;
		rc = _findnexti64(hdir,&dtax)) {
			
			if (skip_this_file(dtax.name,dtax.attrib,ambname,unix_wildcards))
				continue;
				
			foundfile = 1;  /* have found at least one file with pattern */
			++filekt;
			++total_filekt;
					
			parse_name_for_display(dtax.name,full_path,short_name);

			filesize = calculate_filesize(dtax.size,use_blocksize);
			space_used += filesize;                 
			total_space_used += filesize;
			WriteFileInfo(fname,ext,dir,c1,ctl,dtax.attrib,dtax.name,dtax.size,
				dtax.time_access,dtax.time_create,dtax.time_write,
				totals_only,actpath,out);
			if (!totals_only && !nflag)
				putc('\n',out);
			if (showStreams) {
				int rc1;
				char streamname[_MAX_PATH];
				int id;
				char *data;
				char streampath[_MAX_PATH];
				_int64 streamsize;
				rc1 = FirstStream(full_path,streamname,&streamsize,&id);
				while (rc1) {
					if (*streamname) {
						space_used += streamsize;
						total_space_used += streamsize;
						data = strstr(streamname,":$DATA");
						if (data)
							*data = 0;
						strcpy(fname," Stream ");
						sprintf(ext,"%4d",id);
						strcpy(streampath,dtax.name);
						strcat(streampath,streamname);
						WriteFileInfo(fname,ext,dir,c1,ctl,dtax.attrib,streampath,streamsize,
							dtax.time_access,dtax.time_create,dtax.time_write,
							totals_only,actpath,out);
						if (!totals_only && !nflag)
							putc('\n',out);
					}
					rc1 = NextStream(streamname,&streamsize,&id);
				}		
			}
		}
		if (foundfile || totals_only || nflag)
			;
		else
			printf("*** Not Found *** %s\n",searchpath);
	}
	
	if (total_flag) {
		print_totals(prevpath,out,filekt,space_used,total_space,space_avail,actsize);
		if (filekt < total_filekt) {
			sprintf(actsize,"%ld",total_filekt);
			nputs(actsize,0,1);
			nputs(" files, ",0,1);
			sprintf(actsize,"%ld",total_space_used);
			nputs(actsize,0,1);
			nputs(" used.",0,1);
		}
	}
	
	/* reset to old path before looking for next entry */
	/* bdosputc(CPMEOF,out);                */
	exit(0);
}       
